#include <cstdio>
#include <FL/gl.h>
#include "vertex.h"
#include "edge.h"
#include "face.h"

using namespace std;
extern int vertexOffset;
extern int edgeOffset;
 
// Constructor 1
Face::Face():layer(0),facePointLink(-1){
}

// Constructor 2
// edge? is + or -
Face::Face(const int edge1,const int edge2,const int edge3,
	   const int vNum1,const int vNum2,const int vNum3):
  layer(0),facePointLink(-1){
  edgeLink.push_back(edge1);
  edgeLink.push_back(edge2);
  edgeLink.push_back(edge3);
  vertexLink.push_back(vNum1);
  vertexLink.push_back(vNum2);
  vertexLink.push_back(vNum3);  
#ifdef DEBUG
  cout << " three vertices face generated." << endl;
#endif
}

// Constructor 3
// edge? is + or -
Face::Face(const int edge1,const int edge2,const int edge3,const int edge4,
	   const int vNum1,const int vNum2,const int vNum3,const int vNum4):
  layer(0),facePointLink(-1){
  edgeLink.push_back(edge1);
  edgeLink.push_back(edge2);
  edgeLink.push_back(edge3);
  edgeLink.push_back(edge4);
  vertexLink.push_back(vNum1);
  vertexLink.push_back(vNum2);
  vertexLink.push_back(vNum3);
  vertexLink.push_back(vNum4);
#ifdef DEBUG
  cout << "four vertices face generated." << endl;
#endif
}

// Copy Constructor
Face::Face(const Face& rhs):layer(rhs.layer),facePointLink(rhs.facePointLink){
  edgeLink.erase(edgeLink.begin(),edgeLink.end());
  vertexLink.erase(vertexLink.begin(),vertexLink.end());
  for(int cnt1=0;cnt1<rhs.edgeLink.size();cnt1++){
    edgeLink.push_back(rhs.edgeLink[cnt1]);
  }
  for(int cnt1=0;cnt1<rhs.vertexLink.size();cnt1++){
    vertexLink.push_back(rhs.vertexLink[cnt1]);
  }
}

// Destructor
Face::~Face(){
}

// Substitution Operator
const Face&
Face::operator=(const Face& rhs){
  if (this==&rhs)
  return *this;

  layer=rhs.layer;
  //edgeLink=rhs.edgeLink;
  //vertexLink=rhs.vertexLink;
  edgeLink.erase(edgeLink.begin(),edgeLink.end());
  vertexLink.erase(vertexLink.begin(),vertexLink.end());
  for(int cnt1=0;cnt1<rhs.edgeLink.size();cnt1++){
    edgeLink.push_back(rhs.edgeLink[cnt1]);
  }
  for(int cnt1=0;cnt1<rhs.vertexLink.size();cnt1++){
    vertexLink.push_back(rhs.vertexLink[cnt1]);
  }
  facePointLink=rhs.facePointLink;
  return *this;
}

// Calculate Face Point
Vertex
Face::CalcFacePoint(){
  Vertex facePoint;
  //Edge* edgeP;
  for(int cnt1=0;cnt1<edgeLink.size();cnt1++){
    facePoint.Plus(vertex[edge[edgeLink[cnt1]].vertexLink[Left]]);
    facePoint.Plus(vertex[edge[edgeLink[cnt1]].vertexLink[Right]]);
    /*
    edgeP=(Edge*)&(edge[edgeLink[cnt1]]);
    facePoint.Plus(edgeP->ReturnVertex(Right));
    facePoint.Plus(edgeP->ReturnVertex(Left));
    */
  }
  
  facePoint.Multi(1.0/edgeLink.size()/2.0);
  return facePoint;
}

const Vertex&
Face::ReturnFacePoint(){
  return vertex[facePointLink]; 
}

// Print All Elements
void
Face::Print(){
  cout << "edge number : ";
  for(vector<int>::iterator cnt1=edgeLink.begin();cnt1!=edgeLink.end();cnt1++){
    cout << *cnt1 << " ";
  }   
  cout << endl;
}

// Information 
void
Face::Info(){
  Print();
  for(int cnt1=0;cnt1<edgeLink.size();cnt1++){
    cout << "edgeLink[cnt1]==" << edgeLink[cnt1] << endl;
  }
  for(int cnt1=0;cnt1<vertexLink.size();cnt1++){
    cout << "vertexLink[cnt1]==" << vertexLink[cnt1] << endl;
  }
}

const Vector3d
Face::Normal(){
  Vector3d vec1,vec2;

  vec1=vertex[vertexLink[1]];
  vec1.Minus(vertex[vertexLink[0]]);
  vec2=vertex[vertexLink[2]];
  vec2.Minus(vertex[vertexLink[1]]);
  vec2=vec2.Cross(vec1);
  vec2.Normalize();
  return vec2;
}

const Vector3d
Face::Normal(int vNum){
  Vector3d vec1,vec2;
  int v0,v1,v2;
  v0=(vNum+3)%4; // vNum-1+4 
  v1=(vNum+4)%4; // vNum+0+4
  v2=(vNum+5)%4; // vNum+1+4

  vec1=vertex[vertexLink[v1]];
  vec1.Minus(vertex[vertexLink[v0]]);
  vec2=vertex[vertexLink[v2]];
  vec2.Minus(vertex[vertexLink[v1]]);
  vec2=vec2.Cross(vec1);
  vec2.Normalize();
  return vec2;
}


// Draw Face on GL Window
void
Face::Draw(){
  Vector3d vec1,vec2;

  if(vertexLink.size()==3){ // Three vertices Face
    glBegin(GL_TRIANGLES);
    vec1=vertex[vertexLink[1]];
    vec1.Minus(vertex[vertexLink[0]]);
    vec2=vertex[vertexLink[2]];
    vec2.Minus(vertex[vertexLink[1]]);
    vec2=vec2.Cross(vec1);
    vec2.Normalize();

    glNormal3dv(vec2.coord);
    vertex[vertexLink[0]].Draw();
    vertex[vertexLink[1]].Draw();
    vertex[vertexLink[2]].Draw();
    glEnd();
  }else{ // Four vertices Face
    glBegin(GL_TRIANGLE_FAN);
    vec1=vertex[vertexLink[1]];
    vec1.Minus(vertex[vertexLink[0]]);
    vec2=vertex[vertexLink[2]];
    vec2.Minus(vertex[vertexLink[1]]);
    vec2=vec2.Cross(vec1);
    vec2.Normalize();

    glNormal3dv(vec2.coord);
    vertex[vertexLink[0]].Draw();
    vertex[vertexLink[1]].Draw();
    vertex[vertexLink[2]].Draw();
    vertex[vertexLink[3]].Draw();
    glEnd();
  }
}

void
Face::WriteFile(FILE* fp){
  fprintf(fp,"%d %d\n",layer,edgeLink.size());
  // edgeLink[]
  for(int cnt1=0;cnt1<edgeLink.size();cnt1++){
    fprintf(fp,"%d ",edgeLink[cnt1]);
  }
  fprintf(fp,"\n");
  // vertexLink[]
  for(int cnt1=0;cnt1<vertexLink.size();cnt1++){
    fprintf(fp,"%d ",vertexLink[cnt1]);
  }
  fprintf(fp,"\n");
}

void
Face::ReadFile(FILE* fp){
  int edgeLinkSize,tmpEdge;
  fscanf(fp,"%d %d",&layer,&edgeLinkSize);
  // edgePLink
  for(int cnt1=0;cnt1<edgeLinkSize;cnt1++){
    fscanf(fp,"%d",&tmpEdge);
    edgeLink.push_back(tmpEdge+edgeOffset);
  }
  // vertexLink
  for(int cnt1=0;cnt1<edgeLinkSize;cnt1++){
    fscanf(fp,"%d",&tmpEdge);
    vertexLink.push_back(tmpEdge+vertexOffset);
  }
  facePointLink=-1;
}


// return 1 if this face have specified edge
int
Face::HaveEdge(int num){
  for(int cnt1=0;cnt1<edgeLink.size();cnt1++){
    if(edgeLink[cnt1]==num){ // have
      return 1;
    }
  }
  return 0; // not have
}


// This procedure should be called from vertexLink.size()==4 !!!
int
Face::OppositeVertex(int v){
  int o,vertexLinkNumber;
  
  for(int cnt1=0;cnt1<vertexLink.size();cnt1++){
    if(vertexLink[cnt1]==v){
      vertexLinkNumber=cnt1;
      break;
    }
  }
  switch(vertexLinkNumber){
  case 0:
    o=2;
    break;
  case 1:
    o=3;
    break;
  case 2:
    o=0;
    break;
  case 3:
    o=1;
    break;   
  }
  return vertexLink[o];
}

// if vertex[v] has two triangles in this (4 vertices) face,
// return 1. Otherwise return 0.
int
Face::HasBreak(int v){
  if(vertexLink.size()==3){
    return 0;
  }
  int o,vertexLinkNumber;
  for(int cnt1=0;cnt1<vertexLink.size();cnt1++){
    if(vertexLink[cnt1]==v){
      vertexLinkNumber=cnt1;
      break;
    }
  }
  if(vertexLinkNumber==0 || vertexLinkNumber==2){
    return 1;
  }
  return 0;
}


// This returns a index of the edge has vertex[v1] and vertex[v2].
// If hasn't, return -1.
int
Face::FindEdge(int v1,int v2){
  int tmpEdge;
  for(int cnt1=0;cnt1<edgeLink.size();cnt1++){
    tmpEdge=edgeLink[cnt1];
    if((edge[tmpEdge].vertexLink[Left]==v1 &&
	edge[tmpEdge].vertexLink[Right]==v2) ||
       (edge[tmpEdge].vertexLink[Left]==v2 &&
	edge[tmpEdge].vertexLink[Right]==v1)){
      return tmpEdge;
    }
  }
  return -1;
}

void
Face::ReverseNormal(){
  if(vertexLink.size()==3){
    swap(vertexLink[0],vertexLink[1]);
  }else{
    swap(vertexLink[1],vertexLink[3]);
  }
}

