#ifndef _GLWINDOW_H_
#define _GLWINDOW_H_

#include <FL/Fl.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/gl.h>
#include <GL/glu.h>

// definition for DBL_MAX,etc...
#ifdef _WIN32
#include <float.h>
#endif
#ifndef _WIN32
#include <values.h>
#endif

#include "vector3d.h"
#include "vertex.h"
#include "camera.h"

#include "connect.xpm"
#include "trans.xpm"
#include "scale.xpm"
#include "rotate.xpm"
#include "svertex.xpm"
#include "sedge.xpm"
#include "sface.xpm"
#include "extrude.xpm"

void Sync();

void CheckBB(Vertex&);
void MultiMat33(double**,Vector3d,double);
void MakeReshapeList(int,int);
void CalcMat33(double*,Vector3d&,double);
void Identity(double*);


// Button
void SetConnect();
void SetRotate();
void SetScale();
void SetTrans();
void SelectVertex();
void SelectEdge();
void SelectFace();

GLboolean Toggle(GLboolean&);
void ToggleX();
void ToggleY();
void ToggleZ();

Vector3d PassP();

class GLWindow : public Fl_Gl_Window {
private:
  void draw();
  int handle(int);
  void Push();
  void Release();
  void Drag();
  void PreConnect();
  void ReshapeCore();
  int AccuratePick();
  int VaguePick(int);
  int VertexPick();
  int VertexPick2(int);
  int EdgePick();
  int FacePick();
  void CalcCameraPos();
  void DrawAxe();
  void DrawGrid();
  void Display();
  //void ToggleAxeB();

  void DebugBuf(GLenum,GLenum); // for debug

  void StickToGrid(Vector3d&);
  
  Vector3d nowUp; // up vector
  float angleY,angleXZ; 
  float cameraR;
  GLdouble modelMatrix[16], projMatrix[16];
  GLint viewport[4]; // common for Pick
  Vector3d vector1; // for reshape,ete..

  // toggle drawing...
  GLboolean axeFlag;// axe
  GLboolean gridFlag;// grid
  GLboolean vFlag;// vertex
  GLboolean eFlag;// edge
  GLboolean fFlag;// face

  //GLboolean sFlag;// subdivided draw flag

  GLboolean sVFlag;// subdivided vertex 
  GLboolean sEFlag;// subdivided edge
  GLboolean sFFlag;// subdivided face

  
public:
  GLWindow(int, int, int, int, const char* l=0);
  ~GLWindow();
  //const GLWindow& operator=(const GLWindow&);
  void Init();
  void Reshape();

  void SetCamera(int);
  void SelectMode(int,int,int);
  void ConnectMode(int,int,int,GLboolean);
  int ProcessHits(GLint,GLuint*);
  int ProcessHits2(GLint,GLuint*);
  int Key();
  const Vector3d PointOnPlane(Vector3d,Vector3d,int,int);
  void CoordToAngle(Vector3d&,float&,float&);
  void ToggleAxe(GLboolean);
  void ToggleGrid(GLboolean);
  void ToggleVertex(GLboolean);
  void ToggleEdge(GLboolean);
  void ToggleFace(GLboolean);
  void Pivot();
  void Approach();
  void Zoom();
  void Pan();
  void (GLWindow::*MiddleDrag)();
  void (GLWindow::*RightDrag)();
  void (GLWindow::*ShiftRightDrag)();
  void (GLWindow::*AltRightDrag)();
  void SetCameraType();
  
  void ViewTarget();
  void ViewWhole();

  void SVDraw();
  void SEDraw();
  void SFDraw();

  int camNum;
  Camera* camera;
};

#endif // _GLWINDOW_H_
