#include <cstdio>
#include <cstring>
#include <vector>
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "camera.h"
#include "light.h"
#include "pygmalion.h"
#include "operation.h"
#include "usersetting.h"
#include "init.h"
#include "glWindow.h"

// for user setting
char settingFile[512];
char renderExe[512]; // user specify renderer
char renderOpt[512]; // user specify render option
char typeOpt[512]; // auto specific render option
int renderSubTimes=3; // subdivision times for rendering
int pauseCheck=1;
int povType=1;

extern double gridSize;

// bounding box for clipping for drawing by OpenGL
extern double minX,minY,minZ;
extern double maxX,maxY,maxZ;

///////////////////////////////////////////////////
// pygmalion.cc:selection
// index of selected object
extern vector<int> selectV;
extern vector<int> selectE;
extern vector<int> selectF;

extern vector<int> subdividedVertex; 
extern vector<int> subdividedEdge; // Edge
extern vector<int> subdividedFace; // Face
extern int subdivideLevel; // init value=0 mean no subdivision

extern vector<int> stackOP; // stack of operation for UNDOing
extern vector<vector<Vertex> > undoVertex; // undo buffer
extern vector<vector<Edge> > undoEdge;
extern vector<vector<Face> > undoFace;
extern vector<vector<int> > undoSelectV;
extern vector<vector<int> > undoSelectE;
extern vector<vector<int> > undoSelectF;

extern int lastGEOOP; // index in last operation(lastOP reserved when Undo)
extern int lastSELOP; // -1 means no operation
extern int lastCAMOP;
extern int lastTEXOP;
extern int lastOPIndex; // specify last index of operation stack

extern enum {
  uGEOMETRY=1,
  uSELECTION=2,
  uCAMERA=4,
  uTEXTURE=8
}opCode;

extern vector<Camera> cameraList;
extern vector<Light> lightList;

extern UserSetting* userSetting;
extern int precision; // output precision

void
PYGInit(){
  // initialize
  minX=-0.1;minY=-0.1;minZ=-0.1;
  maxX=0.1;maxY=0.1;maxZ=0.1;
  selectV.erase(selectV.begin(),selectV.end());
  selectE.erase(selectE.begin(),selectE.end());
  selectF.erase(selectF.begin(),selectF.end());
  RefineTotalV();
  //strcpy(tmpExpFilename,"tmppygzz.pov");
  // stack
  stackOP.erase(stackOP.begin(),stackOP.end());
  undoVertex.erase(undoVertex.begin(),undoVertex.end());
  undoEdge.erase(undoEdge.begin(),undoEdge.end());
  undoFace.erase(undoFace.begin(),undoFace.end());
  undoSelectV.erase(undoSelectV.begin(),undoSelectV.end());
  undoSelectE.erase(undoSelectE.begin(),undoSelectE.end());
  undoSelectF.erase(undoSelectF.begin(),undoSelectF.end());
  lastGEOOP=-1;
  lastSELOP=-1;
  lastCAMOP=-1;
  lastTEXOP=-1;
  lastOPIndex=-1;
  PushOP(uGEOMETRY|uSELECTION);

  //cameraList.erase(cameraList.begin(),cameraList.end());
  int num=cameraList.size();
  for(int cnt1=0;cnt1<num;cnt1++){
    DeleteView(0); // num-cnt1-1 also will be good
  }

  lightList.erase(lightList.begin(),lightList.end());
  vertex.erase(vertex.begin(),vertex.end());
  edge.erase(edge.begin(),edge.end());
  face.erase(face.begin(),face.end());
  subdividedVertex.erase(subdividedVertex.begin(),subdividedVertex.end());
  subdividedEdge.erase(subdividedEdge.begin(),subdividedEdge.end());
  subdividedFace.erase(subdividedFace.begin(),subdividedFace.end());
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);
  subdivideLevel=0;
  selectV.erase(selectV.begin(),selectV.end());
  selectE.erase(selectE.begin(),selectE.end());
  selectF.erase(selectF.begin(),selectF.end());
  RefineTotalV();
  //removeFace.erase(removeFace.begin(),removeFace.end());
}

void
WriteInit(){
  WriteSetting();
  // write to file
  FILE* fp;
  
  fp=fopen(settingFile,"w");
  fprintf(fp,"# this file is automatically generated.\n");
  fprintf(fp,"# don't edit!\n");

  // Rnderer
  fprintf(fp,"%s\n%s\n","renderExe",renderExe);
  fprintf(fp,"%s\n%s\n","renderOpt",renderOpt);
  fprintf(fp,"%s\n%s\n","typeOpt",typeOpt);
  fprintf(fp,"%s\n%d\n","renderSubTimes",renderSubTimes);
  fprintf(fp,"%s\n%d\n","pauseCheck",pauseCheck);
  fprintf(fp,"%s\n%d\n","povType",povType);  
  fprintf(fp,"%s\n%d\n","precision",precision);

  // Grid
  fprintf(fp,"%s\n%f\n","gridSize",gridSize);
  fclose(fp);
}

void
WriteSetting(){
  // general setting
  // grid setting
  gridSize=atof(userSetting->gridSizeI->value());
  // render setting
  strcpy(renderExe,userSetting->renderExeI->value());
  strcpy(renderOpt,userSetting->renderOptI->value());
  povType=userSetting->povType->value();
  renderSubTimes=atoi(userSetting->subTimesI->value());
  pauseCheck=userSetting->pauseCheck->value();
  // output setting
  precision=atoi(userSetting->precisionI->value());
  Sync();
}

void
ReadInit(){
  // initialize from file
  char tmpPath[255];

  // On windows,usually this case
  // init file is in same directory with exe
  strcat(tmpPath,"./pygmalion.ini");
  if(fopen(tmpPath,"r")){
    strcpy(settingFile,tmpPath); 
  }        
  // users have each init file
  if(getenv("HOME")!=NULL){ // perhaps on unix, !=NULL
    strcpy(tmpPath,getenv("HOME"));
    strcat(tmpPath,"/pygmalion.ini"); // 
    if(fopen(tmpPath,"r")){
      strcpy(settingFile,tmpPath); 
    }
    strcpy(tmpPath,getenv("HOME"));
    strcat(tmpPath,"/.pygmalionrc"); // for unix
    if(fopen(tmpPath,"r")){
      strcpy(settingFile,tmpPath); 
    }        
  }
  if(!strcmp(settingFile,"")){ // set init file
    if(getenv("HOME")!=NULL){
      strcpy(tmpPath,getenv("HOME"));
      strcat(tmpPath,"/.pygmalionrc");
      strcpy(settingFile,tmpPath);
    }
    return;
  }

  // parse file   
  FILE* fp;
  char tmpBuff[255];
  fp=fopen(settingFile,"r");
  while(fgets(tmpBuff,254,fp)){
    if(!strncmp(tmpBuff,"#",1)){ // comment line  -> skip
      continue; 
    }    
    // Rnderer
    if(!strncmp(tmpBuff,"renderExe",9)){ // for ignore CR/LF ??
      fgets(tmpBuff,254,fp);
      strncpy(renderExe,tmpBuff,strlen(tmpBuff)-1); // also
      continue;
    }
    if(!strncmp(tmpBuff,"renderOpt",9)){
      fgets(tmpBuff,254,fp);
      strncpy(renderOpt,tmpBuff,strlen(tmpBuff)-1);
      continue;
    }
    if(!strncmp(tmpBuff,"typeOpt",7)){
      fgets(tmpBuff,254,fp);
      strncpy(typeOpt,tmpBuff,strlen(tmpBuff)-1);
      continue;
    }
    if(!strncmp(tmpBuff,"renderSubTimes",14)){
      fgets(tmpBuff,254,fp);
      renderSubTimes=atoi(tmpBuff);
      continue;
    }
    if(!strncmp(tmpBuff,"pauseCheck",10)){
      fgets(tmpBuff,254,fp);
      pauseCheck=atoi(tmpBuff);
      continue;
    }
    if(!strncmp(tmpBuff,"povType",7)){
      fgets(tmpBuff,254,fp);
      povType=atoi(tmpBuff);
      continue;
    }
    if(!strncmp(tmpBuff,"precision",9)){
      fgets(tmpBuff,254,fp);
      precision=atoi(tmpBuff);
      continue;
    }
    // Grid
    if(!strncmp(tmpBuff,"gridSize",8)){
      fgets(tmpBuff,254,fp);
      gridSize=atof(tmpBuff);
      continue;
    }
  }
  fclose(fp);
}

void
ReadSetting(){
  // Renderer
  userSetting->renderExeI->value(renderExe);
  userSetting->renderOptI->value(renderOpt);
  userSetting->povType->value(povType);
  char tmpBuff[10];
  sprintf(tmpBuff,"%d",renderSubTimes);
  userSetting->subTimesI->value(tmpBuff);
  userSetting->pauseCheck->value(pauseCheck);
  // Output
  sprintf(tmpBuff,"%d",precision);
  userSetting->precisionI->value(tmpBuff);
  // Grid
  sprintf(tmpBuff,"%f",gridSize);
  userSetting->gridSizeI->value(tmpBuff);
}


