#include <algorithm>
#include <vector>
#include <FL/Fl_Menu_Item.H>
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "pygmalion.h"
#include "subdivide.h"
#include "operation.h"
#include "glWindow.h"
#include "mainmenu.h"

using namespace std;

#ifdef DEBUG
#define DebugOut(a); cout << a << endl;
#else
#define DebugOut(a);
#endif

extern MainMenu* mainM;

extern vector<int> subdividedVertex; // Vertex
extern vector<int> subdividedEdge; // Edge
extern vector<int> subdividedFace; // Face

extern vector<int> selectV;
extern vector<int> selectE;
extern vector<int> selectF;

Fl_Menu_Item* undoMA=NULL; // undo menu address
Fl_Menu_Item* redoMA=NULL; // redo menu address

///////////////////////////////////////////////////
// UNDO
vector<int> stackOP; // stack of operation for UNDOing
vector<vector<Vertex> > undoVertex; // undo buffer
vector<vector<Edge> > undoEdge;
vector<vector<Face> > undoFace;
vector<vector<int> > undoSelectV;
vector<vector<int> > undoSelectE;
vector<vector<int> > undoSelectF;

enum {
  uGEOMETRY=1,
  uSELECTION=2,
  uCAMERA=4,
  uTEXTURE=8
}opCode;

int lastGEOOP=-1; // index in last operation(lastOP reserved when Undo)
int lastSELOP=-1; // -1 means no operation
int lastCAMOP=-1;
int lastTEXOP=-1;
int lastOPIndex=-1; // specify last index of operation stack


// called after geometry change,selection,,,
void
PushOP(int op){
  DebugOut("PushOP : lastOPIndex==" << lastOPIndex);
  stackOP.erase(&stackOP[lastOPIndex+1],stackOP.end());
  if(op&uGEOMETRY){
    DebugOut("geometry : lastGEOOP==" << lastGEOOP);
    undoVertex.erase(&undoVertex[lastGEOOP+1],undoVertex.end());
    undoEdge.erase(&undoEdge[lastGEOOP+1],undoEdge.end());
    undoFace.erase(&undoFace[lastGEOOP+1],undoFace.end());
    undoVertex.push_back();
    VectorCopy<Vertex>(vertex,undoVertex.back(),0,subdividedVertex[1]);
    undoEdge.push_back();
    VectorCopy<Edge>(edge,undoEdge.back(),0,subdividedEdge[1]);
    undoFace.push_back();
    VectorCopy<Face>(face,undoFace.back(),0,subdividedFace[1]);
    lastGEOOP++;
  }
  if(op&uSELECTION){
    DebugOut("selection : lastSELOP==" << lastSELOP);
    undoSelectV.erase(&undoSelectV[lastSELOP+1],undoSelectV.end());
    undoSelectE.erase(&undoSelectE[lastSELOP+1],undoSelectE.end());
    undoSelectF.erase(&undoSelectF[lastSELOP+1],undoSelectF.end());
    undoSelectV.push_back();
    VectorCopy<int>(selectV,undoSelectV.back(),0,selectV.size());
    undoSelectE.push_back();
    VectorCopy<int>(selectE,undoSelectE.back(),0,selectE.size());
    undoSelectF.push_back();
    VectorCopy<int>(selectF,undoSelectF.back(),0,selectF.size());
    lastSELOP++;
  }
  if(op&uCAMERA){   
    //cout << "camera" << endl;
  }   
  if(op&uTEXTURE){
    //cout << "texture" << endl;
  }
  stackOP.push_back(op);
  lastOPIndex++;
  if(stackOP.size()>1){
    undoMA->activate();
  }
  redoMA->deactivate();

  mainM->mainWin->redraw();
  //cout << "PushOP after : lastOPIndex==" << lastOPIndex << endl;
  //cout << "stackOP.size()==" << stackOP.size() << endl;

  /*
  for(int cnt1=0;cnt1<undoVertex.size();cnt1++){
    for(int cnt2=0;cnt2<undoVertex[cnt1].size();cnt2++){
      undoVertex[cnt1][cnt2].Print();
    }
    cout << endl;
  }
  cout << endl;
  */
}


// don't call with lastOPIndex<=-1
// so this function should be called by mainM->undoM
void
Undo(){
  //cout << "Undo : lastOPIndex==" << lastOPIndex << endl;
  if((stackOP[lastOPIndex]&uGEOMETRY)){
    //cout << "geometry : lastGEOOP==" << lastGEOOP << endl;
    PreSub();
    VectorCopy<Vertex>(undoVertex[lastGEOOP-1],vertex,0,undoVertex[lastGEOOP-1].size());
    VectorCopy<Edge>(undoEdge[lastGEOOP-1],edge,0,undoEdge[lastGEOOP-1].size());
    VectorCopy<Face>(undoFace[lastGEOOP-1],face,0,undoFace[lastGEOOP-1].size());
    subdividedVertex[1]=vertex.size();
    subdividedEdge[1]=edge.size();
    subdividedFace[1]=face.size();
    PostSub();
    lastGEOOP--;
  }
  if((stackOP[lastOPIndex]&uSELECTION)){
    //cout << "selection : lastSELOP==" << lastSELOP << endl;
    VectorCopy<int>(undoSelectV[lastSELOP-1],selectV,0,undoSelectV[lastSELOP-1].size());    
    VectorCopy<int>(undoSelectF[lastSELOP-1],selectF,0,undoSelectF[lastSELOP-1].size());    
    VectorCopy<int>(undoSelectE[lastSELOP-1],selectE,0,undoSelectE[lastSELOP-1].size());    
    RefineTotalV();
    lastSELOP--;
  }
  if((stackOP[lastOPIndex]&uCAMERA)){
  }
  if((stackOP[lastOPIndex]&uTEXTURE)){
  }
  lastOPIndex--;
  //cout << "Undo after : lastOPIndex==" << lastOPIndex << endl;
  //cout << "stackOP.size()==" << stackOP.size() << endl;
  redoMA->activate();
  if(lastOPIndex==0){ // nothing to undo
    undoMA->deactivate();
  }
  /*
  for(int cnt1=0;cnt1<undoVertex.size();cnt1++){
    for(int cnt2=0;cnt2<undoVertex[cnt1].size();cnt2++){
      undoVertex[cnt1][cnt2].Print();
    }
    cout << endl;
  }
  cout << endl;
  */
  Sync();
}


void
Redo(){
  //cout << "Redo : lastOPIndex==" << lastOPIndex << endl;
  if((stackOP[lastOPIndex+1]&uGEOMETRY)){
    //cout << "geometry" << endl;
    PreSub();
    VectorCopy<Vertex>(undoVertex[lastGEOOP+1],vertex,0,undoVertex[lastGEOOP+1].size());
    VectorCopy<Edge>(undoEdge[lastGEOOP+1],edge,0,undoEdge[lastGEOOP+1].size());
    VectorCopy<Face>(undoFace[lastGEOOP+1],face,0,undoFace[lastGEOOP+1].size());
    subdividedVertex[1]=vertex.size();
    subdividedEdge[1]=edge.size();
    subdividedFace[1]=face.size();
    PostSub();
    lastGEOOP++;
  }
  if((stackOP[lastOPIndex+1]&uSELECTION)){
    //cout << "selection" << endl;
    VectorCopy<int>(undoSelectV[lastSELOP+1],selectV,0,undoSelectV[lastSELOP+1].size());
    VectorCopy<int>(undoSelectE[lastSELOP+1],selectE,0,undoSelectE[lastSELOP+1].size());
    VectorCopy<int>(undoSelectF[lastSELOP+1],selectF,0,undoSelectF[lastSELOP+1].size());
    RefineTotalV();
    lastSELOP++;
  }
  if((stackOP[lastOPIndex+1]&uCAMERA)){
  }
  if((stackOP[lastOPIndex+1]&uTEXTURE)){
  }
  lastOPIndex++;
  undoMA->activate();
  if(lastOPIndex==stackOP.size()-1){ // nothing to undo
    redoMA->deactivate();
  }
  //cout << "Redo after : lastOPIndex==" << lastOPIndex << endl;
  //cout << "stackOP.size()==" << stackOP.size() << endl;
  /*
  for(int cnt1=0;cnt1<undoVertex.size();cnt1++){
    for(int cnt2=0;cnt2<undoVertex[cnt1].size();cnt2++){
      undoVertex[cnt1][cnt2].Print();
    }
    cout << endl;
  }
  cout << endl;
  */
  Sync();
}

// vector copy -- Vertex,Edge,Face
template<typename T>
void VectorCopy(vector<T>& src,vector<T>& des,int begin,int end){
  des.erase(des.begin(),des.end());
  for(int cnt1=begin;cnt1<end;cnt1++){
    des.push_back(src[cnt1]);
  }
}

