#include <iostream>
#include <vector>
#include <cstring>

#include <FL/Fl.H>                      // Main FLTK header file
#include <FL/fl_ask.H>                  // FLTK convenience functions
#include <FL/fl_file_chooser.H>         // FLTK file chooser
#include <Fl/math.h>
#include <Fl/filename.h>

#include "vector3d.h"
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "camera.h"
#include "light.h"
#include "pygmalion.h"
#include "init.h"
#include "subdivide.h"
#include "glWindow.h"
#include "glWindowUI.h"
#include "fileio.h"
#include "operation.h"

#include "cameraset.h"
#include "mainmenu.h"
#include "newFile.h"
#include "usersetting.h"

using namespace std;

#ifdef DEBUG
#define DebugOut(a); cout << a << endl;
#else
#define DebugOut(a);
#endif

/////////////////////////////////////////////
// menu
CameraSet* cameraSet=NULL;
MainMenu* mainM=NULL;
vector<Fl_Menu_Item> menu;
UserSetting* userSetting=NULL;

extern Fl_Menu_Item* undoMA; // undo menu address
extern Fl_Menu_Item* redoMA; // redo menu address
Fl_Menu_Item* restX; // restrict X
Fl_Menu_Item* restY;
Fl_Menu_Item* restZ;


//////////////////////////////////////////////
// needed for file operation
extern char defaultSub[255]; // default subdivide level
int changed = 0;
extern char filename[1024]; // needed in function Load(...)
//char expFilename[1024] = ""; // export file name or temporary

int vertexOffset;  
int edgeOffset;
int faceOffset;


////////////////////////////////////////////////////
// geometry
vector<Vertex> vertex; // Vertex Vector
vector<Edge> edge; // Edge Vector
vector<Face> face; // Face Vector

extern vector<int> subdividedVertex; 
extern vector<int> subdividedEdge; // Edge
extern vector<int> subdividedFace; // Face
extern int subdivideLevel; // init value=0 mean no subdivision
extern int nowLevel; // preserve subdivideLevel

///////////////////////////////////////////////////
// selection
// index of selected object
vector<int> selectV;
vector<int> selectE;
vector<int> selectF;
vector<int> totalV; // total vertex within select{V,E,F} for TransMode

vector<Vector3d> startV; // selected point's start coords;for grid

extern enum {
  sVertex,sEdge,sFace
}sMode; // select target

extern enum {
  uGEOMETRY=1,
  uSELECTION=2,
  uCAMERA=4,
  uTEXTURE=8
}opCode;


/////////////////////////////////////////////////
// mode of left button of mouse
const int sNORMAL=0;
const int sSUBDIVIDE=1;
int nowStatus=sNORMAL;

extern vector<Camera> cameraList;
extern int outputC; // output camera
extern vector<Light> lightList;

vector<GLWindowUI*> glW;

// bounding box for clipping for drawing by OpenGL
extern double minX,minY,minZ;
extern double maxX,maxY,maxZ;
// bounding box of {selected object ,e.g. totalV}
double minXS,maxXS,minYS,maxYS,minZS,maxZS; // {min,max}{X,Y,Z} of selection


void
Unique(vector<int>& vec){
  for(int cnt1=0;cnt1<vec.size();cnt1++){
    int removeNum=vec[cnt1];
    for(int cnt2=cnt1+1;cnt2<vec.size();cnt2++){
      if(vec[cnt2]==removeNum){
	vec.erase(&vec[cnt2]);
	cnt2--;
      }
    }
  }
}

void
RefineTotalV(){
  totalV.erase(totalV.begin(),totalV.end());
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    totalV.push_back(selectV[cnt1]);
  }
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    for(int cnt2=0;cnt2<edge[selectE[cnt1]].vertexLink.size();cnt2++){
      totalV.push_back(edge[selectE[cnt1]].vertexLink[cnt2]);
    }
  }
  for(int cnt1=0;cnt1<selectF.size();cnt1++){
    for(int cnt2=0;cnt2<face[selectF[cnt1]].vertexLink.size();cnt2++){
      totalV.push_back(face[selectF[cnt1]].vertexLink[cnt2]);
    }
  }
  
  // My Unique Algorithm
  Unique(totalV);
  //cout << "totalV" << endl;
  //PrintVec(totalV);
}

void
CheckPointer(){
    for(int cnt1=0;cnt1<glW.size();cnt1++){
/*
	cout << "glW[" << cnt1 << "]=" << glW[cnt1] << endl;
	cout << "glW[" << cnt1 << "]->glWindow=" << glW[cnt1]->glWindow << endl;
	cout << "glW[" << cnt1 << "]->glWindow->camera=" << glW[cnt1]->glWindow->camera << endl;
	cout << "cameraList[" << cnt1 << "]=" << &cameraList[cnt1] << endl;
*/
	if(&cameraList[cnt1]!=glW[cnt1]->glWindow->camera){
	  fl_alert("cameraList[cnt1]!=glW[cnt1]->glWindow->camera !!!");
	}
	for(int cnt2=cnt1+1;cnt2<glW.size();cnt2++){
	  if(glW[cnt1]->glWindow==glW[cnt2]->glWindow){
	    fl_alert("glW[%d] and glW[%d] share same glWindow!!!",cnt1,cnt2);
	  }
	}
	//glW[cnt1]->glWindow
    }
}


void
CheckBBSCore(Vector3d& v){
  // X
  if(v.X()>maxXS){
    maxXS=v.X();
  }
  if(v.X()<minXS){
    minXS=v.X();
  }
  // Y
  if(v.Y()>maxYS){
    maxYS=v.Y();
  }
  if(v.Y()<minYS){
    minYS=v.Y();
  }
  // Z
  if(v.Z()>maxZS){
    maxZS=v.Z();
  }
  if(v.Z()<minZS){
    minZS=v.Z();
  }
}

void
CheckBBS(vector<int>& vec){
  minXS=DBL_MAX/2;maxXS=-1*DBL_MAX/2;
  minYS=DBL_MAX/2;maxYS=-1*DBL_MAX/2;
  minZS=DBL_MAX/2;maxZS=-1*DBL_MAX/2;
  for(int cnt1=0;cnt1<vec.size();cnt1++){
    CheckBBSCore(vertex[vec[cnt1]]);
  }
}

void
CenteringObject(){
  CheckBBS(totalV);
  Vector3d center;
  center.Set((minXS+maxXS)/2,(minYS+maxYS)/2,(minZS+maxZS)/2);    
  for(int cnt1=0;cnt1<totalV.size();cnt1++){
    vertex[totalV[cnt1]].Minus(center);
  }
  Sync();
}

void
AlignViews(){
  int startX,startY;
  int topWin=0;
  startX=mainM->mainWin->x();
  startY=mainM->mainWin->y()+mainM->mainWin->h()+glOffsetB+glOffsetT;
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    if(!glW[cnt1]->mainWindow->visible()){
      continue;
    }
    glW[cnt1]->mainWindow->position(startX,startY);
    startY+=glW[cnt1]->mainWindow->h()+glOffsetT+glOffsetB;
    if(cnt1<glW.size()-1){
      //if(startY+glW[cnt1+1]->mainWindow->h()+glOffsetT+glOffsetB>Fl::h()){
      if(startY+glW[cnt1+1]->mainWindow->h()+glOffsetB>Fl::h()){
	startX+=glW[topWin]->mainWindow->w()+glOffsetR+glOffsetL;
	startY=mainM->mainWin->y()+mainM->mainWin->h()+glOffsetB+glOffsetT;
	topWin=cnt1+1;
      }
    }
  }
}

// POV-Ray's angle means window's width angle,
// but OpenGL's angle means window's height angle.
double
POVAng2OGLAng(Camera& cam,double povAng){
  double oglAng;
  oglAng=atan((double)cam.height/cam.width*tan(povAng/2*M_PI/180))
    /M_PI*180*2.0;
  return oglAng;
}

double
OGLAng2POVAng(Camera& cam,double oglAng){
  double povAng;
  povAng=atan((double)cam.width/cam.height*tan(oglAng/2*M_PI/180))
    /M_PI*180*2.0;
  return povAng;
}

// should be called after creating camera
void
CreateGLWUI(){ 
  // menu modification1
  mainM->AddViewIndex();
  // GLWindouUI initialize
  int cnt=(int)cameraList.size()-1; // >=0
  // retain glW->visible()
  vector<int> viewFlag;
  for(int cnt1=0;cnt1<viewFlag.size();cnt1++){
    viewFlag.push_back(glW[cnt1]->mainWindow->visible());
    //glW[cnt1]->glWindow->hide();
  }
  glW.push_back();
  glW[cnt] = new GLWindowUI;
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    glW[cnt1]->SetCamera(cnt1,cameraList[cnt1].width,cameraList[cnt1].height);
  }
  //glW[cnt]->hide();
  glW[cnt]->show();
  for(int cnt1=0;cnt1<viewFlag.size();cnt1++){
    if(viewFlag[cnt1]){
      glW[cnt1]->show();
    }
  }
  // menu modification2
  mainM->RenewViewIndex();
}

void
CreateView(int type){
  switch(type){
  case -1:
    break;
  default:
    CreateCamera(type);
    CreateGLWUI();
    break;
  }
  CheckPointer();
}

void
PreDeleteView(Fl_Menu_Item*,int num){
  vector<int> viewFlag;
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    viewFlag.push_back(glW[cnt1]->mainWindow->visible());
    glW[cnt1]->hide();
  }
  glW[num]->show();
  if(!fl_ask("Delete this View...")){
    for(int cnt1=0;cnt1<glW.size();cnt1++){
      if(viewFlag[cnt1]){
	glW[cnt1]->show();
      }else{
	glW[cnt1]->hide();
      }
    }
    return;
  }
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    if(viewFlag[cnt1]){
      glW[cnt1]->show();
    }else{
      glW[cnt1]->hide();
    }
  }
  DeleteView(num);
}

void
DeleteView(int num){
  // cameraList
  cameraList.erase(&cameraList[num]);
  // glW
  glW[num]->hide();
  delete glW[num];
  glW.erase(&glW[num]);
  for(int cnt1=0;cnt1<glW.size();cnt1++){
    glW[cnt1]->SetCamera(cnt1,cameraList[cnt1].width,cameraList[cnt1].height);
  }
  // view index
  mainM->DeleteViewIndex(num);
  mainM->RenewViewIndex();
}

void
ToggleView(Fl_Menu_Item*,int num){
  if(glW[num]->mainWindow->visible()){
    glW[num]->hide();
  }else{
    glW[num]->show();
  }
}

void
CreateCamera(int type){
  switch(type){
  case PERSPECTIVE:
    cameraList.push_back(Camera());
    cameraList.back().camera.Set(20,10,90); // Warning! z is inverse 
    cameraList.back().view.Set(0,0,0);
    break;
  case XYPM:
  case XYMP:
    cameraList.push_back(Camera());
    cameraList.back().type=type;
    cameraList.back().camera.Set(0,0,10); // Warning! z is inverse
    break;
  case YZPM:
  case YZMP:
    cameraList.push_back(Camera());
    cameraList.back().type=type;
    cameraList.back().camera.Set(10,0,0); // Warning! z is inverse 
    break;
  case XZPM:
  case XZMP:
    cameraList.push_back(Camera());
    cameraList.back().type=type;
    cameraList.back().camera.Set(0,10,0); // Warning! z is inverse   
    break;
  }
}

void
CreateNewWin(){
   cameraSet->ReadValue(0);  // 0 mean default value
   cameraSet->cameraSetW->set_modal();
   cameraSet->show();  
}

void
ChangeWinProp(){
  DebugOut("ChangeWinProp");
  cameraSet->ReadValue(1);  // 1 mean using camera[outputC]
  cameraSet->cameraSetW->set_modal();
  cameraSet->show();  
}

// Add Edge and Connect Vertex(vNum1_) between vertex(vNum2_)
void
AddEdge(int vNum1_,int vNum2_){
#ifdef DEBUG
  if(0 > vNum1_>=vertex.size() || 0 > vNum2_>=vertex.size()){
    cout << "strange!!!!!!!!!!!! in AddEdge." << endl;
  }
#endif
  if(vNum1_==vNum2_){
    DebugOut("same vertex can't be connected.");
    return;
  }
  // This makes " vNum1 < vNum2 "
  if(vNum1_>vNum2_){
    swap(vNum1_,vNum2_);
  }

  if(nowStatus==sNORMAL){
    // Find pre generated edge.
    int flag=FindEdge((const int)vNum1_,(const int)vNum2_);
    if(flag!=-1){
      DebugOut("These vertices already connected!!!!!!!!!!!!!!!!!!!");
      DebugOut("vNum1,vNum2==" << vNum1_ << "," << vNum2_);
      return;
    }
  }
  edge.push_back(Edge(vNum1_,vNum2_));
  subdividedEdge[subdivideLevel+1]++;
  SetChanged(1);
  vertex[vNum1_].SetEdgeLink();
  vertex[vNum2_].SetEdgeLink();

  // 
  int flag2;
  flag2=edge[edge.size()-1].FindAddFace(); // {three,four} vertices
  // if it subdivides four Edges Face,remove the face.
  if(flag2){
    RemoveExtraFace();
  }
}

void
RemoveExtraFace(){
  int v0,v1;

  for(int cnt1=0;cnt1<face.size();cnt1++){
    if(face[cnt1].edgeLink.size()==4){
      // first vertex pair
      DebugOut("first vertex pair.");
      v0=face[cnt1].vertexLink[0];
      v1=face[cnt1].vertexLink[2];
      if(v0>v1){
	swap(v0,v1);
      }
      if(FindEdge(v0,v1)!=-1){
	DebugOut("Found three Edges Face within four Edges Face.");
	RemoveFace(cnt1);
	cnt1-=1;
	DebugOut("Four edge Face Erased.");
      }else {
	// second vertex
	DebugOut("second vertex pair.");
	v0=face[cnt1].vertexLink[1];
	v1=face[cnt1].vertexLink[3];
	if(v0>v1){
	  swap(v0,v1);
	}
	if(FindEdge(v0,v1)!=-1){
	  DebugOut("Found three Edges Face within four Edges Face.");
	  RemoveFace(cnt1);
	  cnt1-=1;
	  DebugOut("Four Edges Face Erased.");
	}
      }
    }
  }
}

void
RemoveSelected(){
  // Subdivide Level -> 0
  for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
    Subdivide(-1);
  }

  // Face Remove Operation /////////////////////////////////
  // Clearing removeFace
  //removeFace.erase(removeFace.begin(),removeFace.end());
  vector<int> removeFace; // be removed by "delete"
  
  // Making Remove List of Face
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    for(int cnt2=0;cnt2<edge[selectE[cnt1]].facePLink.size();cnt2++){
      removeFace.push_back(edge[selectE[cnt1]].facePLink[cnt2]);
    }
  }
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[selectV[cnt1]].facePLink.size();cnt2++){
      removeFace.push_back(vertex[selectV[cnt1]].facePLink[cnt2]);
    }
  }
#ifdef DEBUG
  cout << "Fat removeFace List." << endl;
  //PrintVec(removeFace);
#endif
  // Erasing extra number
  Unique(removeFace);
#ifdef DEBUG
  cout << "Smart removeFace List." << endl;
  //PrintVec(removeFace);
#endif
  for(int cnt1=0;cnt1<removeFace.size();cnt1++){
    RemoveFace(removeFace[cnt1]);
    for(int cnt2=cnt1;cnt2<removeFace.size();cnt2++){
      if(removeFace[cnt2]>removeFace[cnt1]){
	removeFace[cnt2]-=1;
      }
    }
  }

  // Edge Remove Operation ///////////////////////////////////////
  // Adding Remove List of Edge
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[selectV[cnt1]].edgePLink.size();cnt2++){
      selectE.push_back(vertex[selectV[cnt1]].edgePLink[cnt2]);
    }
  }
  DebugOut("extra selectE");
  // Erasing extra number
  Unique(selectE);
  DebugOut("smart selectE");
  // Erasing Edge entity
  int lV,rV; // Right Vertex,Left Vertex
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    lV=edge[selectE[cnt1]].vertexLink[Left];
    rV=edge[selectE[cnt1]].vertexLink[Right];
    RemoveEdge(selectE[cnt1]);
    for(int cnt2=cnt1;cnt2<selectE.size();cnt2++){ // mod
      if(selectE[cnt2]>selectE[cnt1]){
	selectE[cnt2]-=1;
      }
    }
    Find4Face(lV,rV,cnt1);
  }

  // Vertex Remove Operation /////////////////////////////////////////
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    RemoveVertex(selectV[cnt1]);
    for(int cnt2=cnt1;cnt2<selectV.size();cnt2++){
      if(selectV[cnt2]>selectV[cnt1]){
	selectV[cnt2]-=1;
      }
    }
  }

  selectF.erase(selectF.begin(),selectF.end());  
  selectE.erase(selectE.begin(),selectE.end());
  selectV.erase(selectV.begin(),selectV.end());
  DebugOut("Removed all selected object.");
}

// This Procedure is called only by RemoveExtraFace !
void
RemoveFace(int num){
#ifdef DEBUG
  if(0>num>=face.size()){
    cout << "Strange!!!!!!! in RemoveFace" << endl;
  }
#endif
  for(int cnt1=0;cnt1<face[num].edgeLink.size();cnt1++){
    edge[face[num].edgeLink[cnt1]].UnSetFaceLink(num);
  }
  for(int cnt1=0;cnt1<face[num].vertexLink.size();cnt1++){
    vertex[face[num].vertexLink[cnt1]].UnSetFaceLink(num);
  }

  face.erase(&(face[num]));
  subdividedFace[subdivideLevel+1]--;

  // Modify Face order in facePLink of {Edges,Vertices}
  // vertex
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[cnt1].facePLink.size();cnt2++){
      if(vertex[cnt1].facePLink[cnt2]>num){
	vertex[cnt1].facePLink[cnt2]-=1;
      }
    }
  }
  // edge
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    for(int cnt2=0;cnt2<edge[cnt1].facePLink.size();cnt2++){
      if(edge[cnt1].facePLink[cnt2]>num){
	edge[cnt1].facePLink[cnt2]-=1;
      }
    }
  }
}

// Calling this procedure,
// no Face has num in edgeLink
void
RemoveEdge(int num){
#ifdef DEBUG
  if(0>num>=edge.size()){
    cout << "Strange!!!!!!!!! in RemoveEdge " << endl;
  }
#endif
  for(int cnt1=0;cnt1<edge[num].vertexLink.size();cnt1++){
    vertex[edge[num].vertexLink[cnt1]].UnSetEdgeLink(num);
  }
  edge.erase(&(edge[num]));
  subdividedEdge[subdivideLevel+1]--;

  // Modify Edge order in edgeLink,edgePLink of {Faces,Vertices}
  // Face
  for(int cnt1=0;cnt1<face.size();cnt1++){
    for(int cnt2=0;cnt2<face[cnt1].edgeLink.size();cnt2++){
      if(face[cnt1].edgeLink[cnt2]>num){
	face[cnt1].edgeLink[cnt2]-=1;
      }
    }
  }
  // Vertex
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    for(int cnt2=0;cnt2<vertex[cnt1].edgePLink.size();cnt2++){
      if(vertex[cnt1].edgePLink[cnt2]>num){
	vertex[cnt1].edgePLink[cnt2]-=1;
      }
    }
  }
}

void
RemoveVertex(int num){
#ifdef DEBUG
  if(0>num>=vertex.size()){
    cout << "Strange!!!!!!!!! in RemoveVertex" << endl;
  }
#endif
  vertex.erase(&(vertex[num]));
  subdividedVertex[subdivideLevel+1]--;

  // Modify Face order in facePLink of Edges
  // edge
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    for(int cnt2=0;cnt2<edge[cnt1].vertexLink.size();cnt2++){
      if(edge[cnt1].vertexLink[cnt2]>num){
	edge[cnt1].vertexLink[cnt2]-=1;
      }
    }
  }

  for(int cnt1=0;cnt1<face.size();cnt1++){
    for(int cnt2=0;cnt2<face[cnt1].vertexLink.size();cnt2++){
      if(face[cnt1].vertexLink[cnt2]>num){
	face[cnt1].vertexLink[cnt2]-=1;
      }
    }
  }
}

// condition : edge1 < edge2
//inline
int
FindEdge(const int v1,const int v2){
  int cnt1;
#ifdef DEBUG
  if(0>v1>=vertex.size()||0>v2>=vertex.size()){
    cout << "strange!!!!!!! in FindEdge." << endl;    
  }
#endif
  int endCnt=edge.size();
  for(cnt1=0;cnt1<endCnt;cnt1++){
    if(edge[cnt1].vertexLink[0] == v1){
      if(edge[cnt1].vertexLink[1] == v2){
	return cnt1;
      }
    }
  }

  return -1;
}

void
Sort3(int& eNum1_,int& eNum2_,int& eNum3_){
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
}

// This procedure should be called under ordered + or - eNum? 
void
AddFace(int eNum1_,int eNum2_,int eNum3_,int vNum1,int vNum2,int vNum3){
  // To do {eNum1_ < eNum2_ < eNum3_}
  Sort3(eNum1_,eNum2_,eNum3_);

  face.push_back(Face(eNum1_,eNum2_,eNum3_,vNum1,vNum2,vNum3));
  subdividedFace[subdivideLevel+1]++;
  SetChanged(1);

  // This make Connected Vertices known their Face.
  vertex[vNum1].SetFaceLink(); 
  vertex[vNum2].SetFaceLink(); 
  vertex[vNum3].SetFaceLink();   
  
  edge[eNum1_].SetFaceLink();
  edge[eNum2_].SetFaceLink();
  edge[eNum3_].SetFaceLink();
}


// Add Face and Connect Edge{eNum1_,eNum2_,eNum3_,eNum4}
void
AddFace(int eNum1_,int eNum2_,int eNum3_,int eNum4_,
	int vNum1,int vNum2,int vNum3,int vNum4){
  // To do {eNum1_ < eNum2_ < eNum3_ < eNum4_}
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum3_>eNum4_){
    swap(eNum3_,eNum4_);
  }
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum3_>eNum4_){
    swap(eNum3_,eNum4_);
  }
  if(eNum1_>eNum2_){
    swap(eNum1_,eNum2_);
  }
  if(eNum2_>eNum3_){
    swap(eNum2_,eNum3_);
  }
  if(eNum3_>eNum4_){
    swap(eNum3_,eNum4_);
  }
  //4 3 2 1 -> 3 4 2 1 -> 3 2 4 1 -> 3 2 1 4 ->
  //2 3 1 4 -> 2 1 3 4 -> 1 2 3 4 

  face.push_back(Face(eNum1_,eNum2_,eNum3_,eNum4_,vNum1,vNum2,vNum3,vNum4));
  subdividedFace[subdivideLevel+1]++;
  SetChanged(1);
  
  vertex[vNum1].SetFaceLink();
  vertex[vNum2].SetFaceLink();
  vertex[vNum3].SetFaceLink();
  vertex[vNum4].SetFaceLink();

  edge[eNum1_].SetFaceLink();
  edge[eNum2_].SetFaceLink();
  edge[eNum3_].SetFaceLink();
  edge[eNum4_].SetFaceLink();
}

void
Find4Face(int lV,int rV,int edgeNum){
  Vertex* v1=&vertex[lV];
  Vertex* v2=&vertex[rV];
  vector<int> connectEdge1;
  vector<int> connectEdge2;
  vector<int> connectV;
  int awayV1,awayV2;
  
  for(int cnt1=0;cnt1<v1->edgePLink.size();cnt1++){
    edge[v1->edgePLink[cnt1]];
    awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Left];
    if(awayV1==lV){
      awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Right];
    }
    for(int cnt2=0;cnt2<v2->edgePLink.size();cnt2++){
      edge[v2->edgePLink[cnt2]];
      awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Left];
      if(awayV2==rV){
	awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Right];
      }      
      if(awayV1==awayV2){ // found connected !
	DebugOut("found connected! in Find4Face.");
	connectEdge1.push_back(v1->edgePLink[cnt1]);
	connectEdge2.push_back(v2->edgePLink[cnt2]);
	connectV.push_back(awayV1);
      }
    }
  }
#ifdef DEBUG
  cout << "connectEdge1" << endl;
  PrintVec(connectEdge1);
  cout << "connectEdge2" << endl;
  PrintVec(connectEdge2);
  cout << "connectV" << endl;
  PrintVec(connectV);
  cout << "selectE" << endl;
  PrintVec(selectE);
#endif
  
  int flag;
  for(int cnt1=0;cnt1<connectV.size();cnt1++){
    for(int cnt2=cnt1+1;cnt2<connectV.size();cnt2++){
      flag=0;

      for(int cnt3=edgeNum+1;cnt3<selectE.size();cnt3++){
	if(connectEdge1[cnt1]==selectE[cnt3]){
	  flag=1;
	}
	if(connectEdge1[cnt2]==selectE[cnt3]){
	  flag=1;
	}
	if(connectEdge2[cnt1]==selectE[cnt3]){
	  flag=1;
	}
	if(connectEdge2[cnt2]==selectE[cnt3]){
	  flag=1;
	}
      }
      DebugOut("flag==" << flag);
      if(flag==1){
	continue;
      }
      AddFace(connectEdge1[cnt1],connectEdge1[cnt2],
	      connectEdge2[cnt1],connectEdge2[cnt2],
	      lV,connectV[cnt1],rV,connectV[cnt2]);
    }
  }
}

void
PrintHelp(){
  cout << "Usage:" << endl;
  cout << "'s'     toggle [subdivision surface] or [polygon]  ." << endl;
  cout << "" << endl;
  cout << "" << endl;
}

void
SelectAll(){
  switch(sMode){
  case sVertex:
    selectV.erase(selectV.begin(),selectV.end());
    for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
      selectV.push_back(cnt1);
    }
    break;
  case sEdge:
    selectE.erase(selectE.begin(),selectE.end());
    for(int cnt1=0;cnt1<subdividedEdge[1];cnt1++){
      selectE.push_back(cnt1);
    }
    break;
  case sFace:
    selectF.erase(selectF.begin(),selectF.end());
    for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
      selectF.push_back(cnt1);
    }
    break;
  }
  RefineTotalV();
}

void
DeselectAll(){
  switch(sMode){
  case sVertex:
    selectV.erase(selectV.begin(),selectV.end());
    break;
  case sEdge:      
    selectE.erase(selectE.begin(),selectE.end());
    break;
  case sFace:
    selectF.erase(selectF.begin(),selectF.end());      
    break;
  }
  RefineTotalV();
}

void
DeselectAllofAll(){
  selectV.erase(selectV.begin(),selectV.end());
  selectE.erase(selectE.begin(),selectE.end());
  selectF.erase(selectF.begin(),selectF.end());      
  RefineTotalV();
}

void
Extrude(){
  PreSub();
  ExtrudeCore();
  PostSub();
  Sync();
  PushOP(uGEOMETRY|uSELECTION);
}

void
ExtrudeCore(){

  int nowVertexSize=vertex.size();
  int nowSelectVSize=selectV.size();
  int edge1,edge2;
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    vertex.push_back(Vertex(0,0,0));
    subdividedVertex[1]++;
    SetChanged(1);
    vertex[vertex.size()-1].SetV(vertex[selectV[cnt1]]);    
  }
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    edge1=selectV[cnt1];
    for(int cnt2=0;cnt2<selectV.size();cnt2++){
      edge2=selectV[cnt2];
      DebugOut("edge1,edge2:" << edge1 << "," << edge2);
      if(FindEdge(edge1,edge2)!=-1){ // found 
	DebugOut("found edge");
	AddEdge(nowVertexSize+cnt1,nowVertexSize+cnt2);
      }
    }
  }
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    AddEdge(selectV[cnt1],nowVertexSize+cnt1);
  }
  selectV.erase(selectV.begin(),selectV.end());
  for(int cnt1=0;cnt1<nowSelectVSize;cnt1++){
    selectV.push_back(nowVertexSize+cnt1);
  }
  RefineTotalV();
}


void
FourToThree(){
  for(int cnt1=subdividedFace[0];cnt1<subdividedFace[1];cnt1++){
    if(face[cnt1].vertexLink.size()==4){
      AddEdge(face[cnt1].vertexLink[0],face[cnt1].vertexLink[2]);
      cnt1--;
    }
  }
}


void
DebugObjects(){
  cout << endl << "In DebugObjects" << endl;
  cout << "subFace " << endl;
  for(int cnt1=0;cnt1<subdividedFace.size();cnt1++){
    cout << subdividedFace[cnt1] << endl;
  }
  cout << "end" << endl;

  cout << "subEdge " << endl;
  for(int cnt1=0;cnt1<subdividedEdge.size();cnt1++){
    cout << subdividedEdge[cnt1] << endl;
  }
  cout << "end" << endl;

  cout << "subVertex " << endl;
  for(int cnt1=0;cnt1<subdividedVertex.size();cnt1++){
    cout << subdividedVertex[cnt1] << endl;
  }
  cout << "end" << endl;

  for(int cnt1=0;cnt1<face.size();cnt1++){
    cout << "face[" << cnt1 << "]" << endl;
    face[cnt1].Info();
  }
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    cout << "edge[" << cnt1 << "]" << endl;
    edge[cnt1].Info();
  }
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    cout << "vertex[" << cnt1 << "]" << endl;
    vertex[cnt1].Info();
  }
}


void
SelectedSharp(){
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    vertex[selectV[cnt1]].MoreSharp();
  }
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    edge[selectE[cnt1]].MoreSharp();
  }
}

void
SelectedSmooth(){
  for(int cnt1=0;cnt1<selectV.size();cnt1++){
    vertex[selectV[cnt1]].MoreSmooth();
  }
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    edge[selectE[cnt1]].MoreSmooth();
  }
}

void
WeightPropagate(){
  for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
    for(int cnt2=0;cnt2<subdividedEdge[cnt1];cnt2++){
      vertex[edge[cnt2].edgePointLink].weight=
	(vertex[edge[cnt2].vertexLink[Left]].weight+
	 vertex[edge[cnt2].vertexLink[Right]].weight)/2;
    }
  }
}


// Print All Edge
void
PrintEdge(){
  for(int cnt1=0;cnt1<edge.size();cnt1++){
    edge[cnt1].Print();
  }
  cout << "edge num:" << edge.size() << endl;
}

// Print All Vertex
void
PrintVertex(){
  for(int cnt1=0;cnt1<vertex.size();cnt1++){
    vertex[cnt1].Print();
  }
  cout << "vertex num:" << vertex.size() << endl;
}

// Print All Face
void
PrintFace(){
  for(int cnt1=0;cnt1<face.size();cnt1++){
    face[cnt1].Print();
  }
  cout << "face num:" << face.size() << endl;
}


void
PrintVec(vector<int> vec){
  cout << "print vec" << endl;
  for(int cnt1=0;cnt1<vec.size();cnt1++){
    cout << vec[cnt1] << endl;
  }
}

int
CheckSave(void) {
  if (!changed) return 1;

  int choice;
  choice=fl_choice("Current file has not been saved.\n"
		   "Would you like to save it now?",
		   "Cancel","Yes","No");
  switch(choice){
  case 0:
    return (0);
    break;
  case 1:
    SaveCB();
    return (1);
    break;
  case 2:
    return (1);
    break;
  }
}

void
NewCB(void) {
  if (changed)
    if (!CheckSave()) return;

  NewFile fileDialog;
  fileDialog.mainWin->set_modal();
  fileDialog.show();
  for (;;) {
    Fl::wait();
    Fl_Widget *o;
    while ((o = Fl::readqueue())) {
      if (o == fileDialog.ok) {
	filename[0] = '\0';
	SetChanged(1);
	SetChanged(0);
	PYGInit();
	fileDialog.CreateAllView();
	fileDialog.mainWin->hide();
	AlignViews();
	Fl::redraw();
	return;
      } else if (o == fileDialog.cancel || o == fileDialog.mainWin) {
	fileDialog.mainWin->hide();
	return;
      }
    }
  }
}

void
OpenCB(void) {
  char *newfile;

  if (changed)
    if (!CheckSave()) return;

  DebugOut("In OpenCB;filename==" << filename);
  newfile = fl_file_chooser("Open File?", "*.pyg", filename);
  if (newfile != NULL) Load(newfile,OPEN);
}

void
ChangedCB(void) {
  SetChanged(1);
}

void
SetChanged(int c) {
  if (c != changed) {
    char title[1024];
    char *slash;

    changed = c;

    strcpy(title,"Pygmalion:");
    if (filename[0] == '\0') strcat(title, "Untitled.pyg");
    else {
      slash = strrchr(filename, '/');
      if (slash == NULL) slash = strrchr(filename, '\\');

      if (slash != NULL) strcat(title, slash + 1);
      else strcat(title, filename);
    }
    if (changed) strcat(title, " (modified)");
    mainM->mainWin->label(title);
  }
}

void
SaveCB(void) {
  if (filename[0] == '\0') {
    // No filename - get one!
    SaveAsCB();
    return;
  } else {
    DebugOut("In SaveCB;filename==" << filename);
    Save(filename);
  }
}

void
SaveAsCB(void) {
  char *newfile;

  newfile = fl_file_chooser("Save File As?", "*.pyg", filename);
  if (newfile != NULL) Save(newfile);
}

void
QuitCB(void) {
  if (changed){
    if (!CheckSave()){
      return;
    }
  }
  
  mainM->mainWin->hide();  

  for(int cnt1=0;cnt1<glW.size();cnt1++){
    glW[cnt1]->hide();
    delete glW[cnt1];
    glW[cnt1]=NULL;
  }
  
  delete cameraSet;
  cameraSet=NULL;
  delete mainM;
  mainM=NULL;
  
  exit(0);
}

void
CancelCB(){
  return;
}


// Test Main Procedure
int
main(int argc,char** argv){
  // always subdividedNumber{V,E,F}[0]==0
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);

  // number of objects in subdivide level 0 
  subdividedVertex.push_back(0);
  subdividedEdge.push_back(0);
  subdividedFace.push_back(0);
  sprintf(defaultSub,"3"); // may overwrite in ReadInit()
  ReadInit();
  mainM = new MainMenu;

  undoMA=mainM->undoM;
  redoMA=mainM->redoM;
  restX=mainM->restrictXM;
  restY=mainM->restrictYM;
  restZ=mainM->restrictZM;

  mainM->mainWin->show();
  //mainM->mainWin->position(0,0);

  cameraSet = new CameraSet; // camera property dialog
  userSetting = new UserSetting; // User Setting dialog

  PYGInit();

  // if given input file,
  if (argc > 1) Load(argv[1],OPEN);  

  Fl::run();

  return 0;
}


// arrange all face's normal direction
// if edge[].facePLink.size()>2, stop and return -1;
// normally return 0;
int
ArrangeNormal(){
  if(face.size()==0){
    return 0;
  }
  vector<int> check;
  for(int cnt1=0;cnt1<face.size();cnt1++){
    check.push_back(0);
  }
  vector<int> faceStack;

  // start face
  int start=0; // start of each propagation
  int now=0; // top of FIFO stack
  
  while(start<check.size()){
    while(check[start]!=0 && start<check.size()){
      start++;
    }
    if(start==check.size()){
      return 0; // end
    }
    faceStack.push_back(start);
    while(faceStack.size()!=0){
      now=faceStack.front();
      check[now]=1;
      // propagate face[now]'s normal direction
      for(int cnt2=0;cnt2<face[now].edgeLink.size();cnt2++){
	// adjacent face
	int adjacentFace;
	if(edge[face[now].edgeLink[cnt2]].facePLink.size()==2){     
	  if(edge[face[now].edgeLink[cnt2]].facePLink[0]==now){
	    adjacentFace=edge[face[now].edgeLink[cnt2]].facePLink[1];
	  }else{
	    adjacentFace=edge[face[now].edgeLink[cnt2]].facePLink[0];
	  }
	}else{ // boundary edge or 3,or more facePLink
	  continue;
	}
	
	// insert 
	if(check[adjacentFace]==0){
	  faceStack.push_back(adjacentFace);
	}else{
	  continue;
	}

	// check whether both now face's normal and adjacent face's one are same directional.	
	int normalCheck=SameDirection(face[now].edgeLink[cnt2]);
	switch(normalCheck){
	case -1:
	  return -1;
	  break;
	case 0:
	  break;
	case 1:
	  face[adjacentFace].ReverseNormal();
	  break;
	}
	check[adjacentFace]=1;
      }
      // delete first element of FIFO stack
      faceStack.erase(faceStack.begin());
      //check[now]=1;
    }
  }
  PushOP(uGEOMETRY);
}
  
void
ReverseAllNormal(){
  for(int cnt1=0;cnt1<face.size();cnt1++){
    face[cnt1].ReverseNormal();
  }
}

// this function returns ;
// -1 in case : edge[].facePLink.size()>2 
// 0 in case : edge[].facePLink.size()==1 or ==2,2-manifold, have same direction normals
// 1 in case : edge[],facePLink.size()==2, have different direction normals
int
SameDirection(int edge1){
  if(edge[edge1].facePLink.size()>2){
    return -1;
  }
  if(edge[edge1].facePLink.size()==1){
    return 0;
  }
  int v1,v2;
  v1=edge[edge1].vertexLink[0];
  v2=edge[edge1].vertexLink[1];  
  int f1,f2;
  f1=edge[edge1].facePLink[0];
  f2=edge[edge1].facePLink[1];
  // find v1 in 2face's vertexLink
  for(int cnt1=0;cnt1<face[f1].vertexLink.size();cnt1++){
    if(face[f1].vertexLink[cnt1]==v1){
      for(int cnt2=0;cnt2<face[f2].vertexLink.size();cnt2++){
	if(face[f2].vertexLink[cnt2]==v1){
	  // find v2 in 2face's vertexLink
	  int preV1,postV1;
	  if(cnt1==face[f1].vertexLink.size()-1){
	    postV1=face[f1].vertexLink[0];
	  }else{
	    postV1=face[f1].vertexLink[cnt1+1];
	  }
	  if(cnt1==0){
	    preV1=face[f1].vertexLink.back();
	  }else{
	    preV1=face[f1].vertexLink[cnt1-1];
	  }

	  int preV2,postV2;
	  if(cnt2==face[f2].vertexLink.size()-1){
	    postV2=face[f2].vertexLink[0];
	  }else{
	    postV2=face[f2].vertexLink[cnt2+1];
	  }
	  if(cnt2==0){
	    preV2=face[f2].vertexLink.back();
	  }else{
	    preV2=face[f2].vertexLink[cnt2-1];
	  }
	  
	  if(preV1==v2){
	    if(preV2==v2){
	      return 1; // different direction
	    }else{// postV2==v2
	      return 0; // same direction
	    }
	  }else{// postV1==v2
	    if(preV2==v2){
	      return 0; // same direction
	    }else{// postV2==v2
	      return 1; // different direction
	    }
	  }
	}
      }
    }
  }
}

