#include <ctime>
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "pygmalion.h"
#include "subdivide.h"

#ifdef DEBUG
#define DebugOut(a); cout << a << endl;
#else
#define DebugOut(a);
#endif

int subdivideLevel=0; // init value=0 mean no subdivision
int nowLevel; // preserve subdivideLevel
char defaultSub[255]; // default subdivide level

////////////////////////////////////////////////////
// geometry
extern vector<Vertex> vertex; // Vertex Vector
extern vector<Edge> edge; // Edge Vector
extern vector<Face> face; // Face Vector

// order of Vertex in subdivide level
// ex.{10,20,45} -> Level 1 {0-9}, Level 2 {10-19}, Level 3 {20-44}
vector<int> subdividedVertex; 
vector<int> subdividedEdge; // Edge
vector<int> subdividedFace; // Face

///////////////////////////////////////////////////
// selection
// index of selected object
extern vector<int> selectV;
extern vector<int> selectE;
extern vector<int> selectF;
extern vector<int> totalV; // total vertex within select{V,E,F} for TransMode

void
SubdivideCore(int fStart,int fEnd,int eStart,int eEnd,int vStart,int vEnd){
  // Add Calculated Point
  
  // Face
  for(int cnt1=fStart;cnt1<fEnd;cnt1++){
    DebugOut("Face[" << cnt1 << "]");
    vertex.push_back(Vertex(face[cnt1].CalcFacePoint()));	
    subdividedVertex[subdivideLevel+1]++;
    SetChanged(1);
    face[cnt1].facePointLink=vertex.size()-1;
  }
  
  // Edge
  for(int cnt2=eStart;cnt2<eEnd;cnt2++){
    DebugOut("Edge[" << cnt2 << "]");
    vertex.push_back(Vertex(edge[cnt2].CalcEdgePoint()));
    subdividedVertex[subdivideLevel+1]++;
    SetChanged(1);
    edge[cnt2].edgePointLink=vertex.size()-1;
    edge[cnt2].ConnectToFacePoint();
  }
  
  // Vertex
  for(int cnt3=vStart;cnt3<vEnd;cnt3++){
    DebugOut("Vertex[" << cnt3 << "]");
    if(vertex[cnt3].facePLink.size()>=1){
      vertex.push_back(Vertex(vertex[cnt3].CalcVertexPoint()));
      subdividedVertex[subdivideLevel+1]++;
      SetChanged(1);
      vertex[cnt3].vertexPointLink=vertex.size()-1;
      vertex[cnt3].ConnectToEdgePoint();
    }
  }      
}

// levelTrans==+1 --> higher subdivide level display mode
// levelTrans==-1 --> lower subdivide level display mode
// levelTrans== 0 --> once subdivide
void
Subdivide(int levelTrans){  

  if(face.size()==0){
    DebugOut("Face not found.");
    return;
  }

  time_t startTime;
  time_t endTime;
  double workTime;
  
  int faceStart,faceEnd;
  int edgeStart,edgeEnd;
  int vertexStart,vertexEnd;

  // Time 
  startTime=time(NULL);

  switch(levelTrans){   
    
  case -1: // lower subdivide level ////////////////////////////
    if(subdivideLevel==0){
      DebugOut("now subdivide level is 0.");
      return;
    }

    // Face
    face.erase(&face[subdividedFace[subdivideLevel]],
	       &face[subdividedFace[subdivideLevel+1]]);
    //subdividedFace.pop_back();
    subdividedFace.erase(subdividedFace.end()-1);

    // Edge
    edge.erase(&edge[subdividedEdge[subdivideLevel]],
	       &edge[subdividedEdge[subdivideLevel+1]]);
    //subdividedEdge.pop_back();
    subdividedEdge.erase(subdividedEdge.end()-1);

    // Vertex
    vertex.erase(&vertex[subdividedVertex[subdivideLevel]],
		 &vertex[subdividedVertex[subdivideLevel+1]]);
    //subdividedVertex.pop_back();
    subdividedVertex.erase(subdividedVertex.end()-1);

    // change {face,edge,vertex}PointLink -> -1
    for(int cnt1=subdividedVertex[subdivideLevel-1];
	cnt1<subdividedVertex[subdivideLevel];cnt1++){
      vertex[cnt1].vertexPointLink=-1;
    }
    for(int cnt1=subdividedEdge[subdivideLevel-1];
	cnt1<subdividedEdge[subdivideLevel];cnt1++){
      edge[cnt1].edgePointLink=-1;
    }
    for(int cnt1=subdividedFace[subdivideLevel-1];
	cnt1<subdividedFace[subdivideLevel];cnt1++){
      face[cnt1].facePointLink=-1;
    }
    
    subdivideLevel-=1;
    DebugOut("now subdivide level is " << subdivideLevel);
    break;
    
  case +1: // higher level subdivide level /////////////////////////

    nowStatus=sSUBDIVIDE;

    // important !!!!!!!!
    // preallocation preserve validating iterator !! 
    vertex.reserve(2*vertex.size()+edge.size()+face.size());
    edge.reserve(edge.size()+2*edge.size()+face.size()*4);
    face.reserve(face.size()*4);
    
    // OK
    faceStart=subdividedFace[subdivideLevel];
    faceEnd=subdividedFace[subdivideLevel+1];
    edgeStart=subdividedEdge[subdivideLevel];
    edgeEnd=subdividedEdge[subdivideLevel+1];
    vertexStart=subdividedVertex[subdivideLevel];
    vertexEnd=subdividedVertex[subdivideLevel+1];

    subdividedVertex.push_back(vertexEnd);
    subdividedEdge.push_back(edgeEnd);
    subdividedFace.push_back(faceEnd);

    subdivideLevel+=1;    
    SubdivideCore(faceStart,faceEnd,edgeStart,edgeEnd,vertexStart,vertexEnd);

    DebugOut("now subdivide level is " << subdivideLevel);
    nowStatus=sNORMAL;
    break;
    
  case 0: // once subdivide  /////////////////////////////////////

    selectF.erase(selectF.begin(),selectF.end());  
    selectE.erase(selectE.begin(),selectE.end());
    selectV.erase(selectV.begin(),selectV.end());
    RefineTotalV();

    if(subdivideLevel==0){ // need subdivide
      nowStatus=sSUBDIVIDE;
      Subdivide(1);
      Subdivide(0);

      DebugOut("now subdivide level is " << subdivideLevel);
      nowStatus=sNORMAL;

    }else { // subdivideLevel>0 // need no subdivide 
      face.erase(&(face[0]),&(face[subdividedFace[1]]));
      edge.erase(&(edge[0]),&(edge[subdividedEdge[1]]));
      vertex.erase(&(vertex[0]),&(vertex[subdividedVertex[1]]));      

      // fix links in objects
      for(int cnt1=0;cnt1<face.size();cnt1++){
	for(int cnt2=0;cnt2<face[cnt1].edgeLink.size();cnt2++){
	  face[cnt1].edgeLink[cnt2]-=subdividedEdge[1];
	}
	for(int cnt2=0;cnt2<face[cnt1].vertexLink.size();cnt2++){
	  face[cnt1].vertexLink[cnt2]-=subdividedVertex[1];
	}
	if(face[cnt1].facePointLink!=-1){
	  face[cnt1].facePointLink-=subdividedVertex[1];
	}
      }


      for(int cnt1=0;cnt1<edge.size();cnt1++){
	for(int cnt2=0;cnt2<edge[cnt1].vertexLink.size();cnt2++){
	  edge[cnt1].vertexLink[cnt2]-=subdividedVertex[1];
	}
	for(int cnt2=0;cnt2<edge[cnt1].facePLink.size();cnt2++){
	  edge[cnt1].facePLink[cnt2]-=subdividedFace[1];
	}
	if(edge[cnt1].edgePointLink!=-1){
	  edge[cnt1].edgePointLink-=subdividedVertex[1];
	}
      }

      for(int cnt1=0;cnt1<vertex.size();cnt1++){
	for(int cnt2=0;cnt2<vertex[cnt1].edgePLink.size();cnt2++){
	  vertex[cnt1].edgePLink[cnt2]-=subdividedEdge[1];
	}
	for(int cnt2=0;cnt2<vertex[cnt1].facePLink.size();cnt2++){
	  vertex[cnt1].facePLink[cnt2]-=subdividedFace[1];
	}
	if(vertex[cnt1].vertexPointLink!=-1){
	  vertex[cnt1].vertexPointLink-=subdividedVertex[1];
	}
      }      

      // fix objects number in sudividedNumber{V,E,F}
      // dont need cnt1=0 
      for(int cnt1=subdivideLevel+1;cnt1>0;cnt1--){
	subdividedFace[cnt1]-=subdividedFace[1];
	subdividedEdge[cnt1]-=subdividedEdge[1];
	subdividedVertex[cnt1]-=subdividedVertex[1];
      }

      subdividedFace.erase(subdividedFace.begin());
      subdividedEdge.erase(subdividedEdge.begin());
      subdividedVertex.erase(subdividedVertex.begin());
      
      subdivideLevel-=1;
      
      DebugOut("now subdivide level is " << subdivideLevel);
    } 
    break;
  }
  
  endTime=time(NULL);
  workTime=difftime(endTime,startTime);  
  DebugOut("Subdivide time : " << workTime);
}

void
PreSub(){
  nowLevel=subdivideLevel;
  for(int cnt1=0;cnt1<nowLevel;cnt1++){
    Subdivide(-1);
  }
}

void
PostSub(){
  for(int cnt1=0;cnt1<nowLevel;cnt1++){
    Subdivide(1);
  }
}

void
RecalcSub(int level){ // level is recalculating level now
  int faceS,faceE;
  int edgeS,edgeE;
  int vertexS,vertexE;

  Vertex tmpV;

  faceS=subdividedFace[level];
  faceE=subdividedFace[level+1];
  edgeS=subdividedEdge[level];
  edgeE=subdividedEdge[level+1];
  vertexS=subdividedVertex[level];
  vertexE=subdividedVertex[level+1];

  for(int cnt1=faceS;cnt1<faceE;cnt1++){
    tmpV=face[cnt1].CalcFacePoint();
    vertex[face[cnt1].facePointLink].SetV(tmpV);
  }
  for(int cnt1=edgeS;cnt1<edgeE;cnt1++){
    tmpV=edge[cnt1].CalcEdgePoint();
    vertex[edge[cnt1].edgePointLink].SetV(tmpV);
  }
  for(int cnt1=vertexS;cnt1<vertexE;cnt1++){
    if(vertex[cnt1].vertexPointLink!=-1){
      tmpV=vertex[cnt1].CalcVertexPoint();
      vertex[vertex[cnt1].vertexPointLink].SetV(tmpV);
    }
  }
}
