#include <cstdio>
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "pygmalion.h"
#include <FL/gl.h>
 
using namespace std;

extern vector<Edge> edge;
extern vector<Face> face;

extern double minX,minY,minZ;
extern double maxX,maxY,maxZ;
//extern int vertexOffset;
extern int edgeOffset;
extern int faceOffset;
  
// Constructor
Vertex::Vertex():Vector3d(),vertexPointLink(-1),weight(0.5){
}

// Constructor
Vertex::Vertex(const double& d1,const double& d2,const double& d3):
  Vector3d(d1,d2,d3),vertexPointLink(-1),weight(0.5){
}

// Copy Constructor
Vertex::Vertex(const Vertex& rhs):Vector3d(rhs),
    vertexPointLink(rhs.vertexPointLink),weight(rhs.weight){
  facePLink.erase(facePLink.begin(),facePLink.end());
  edgePLink.erase(edgePLink.begin(),edgePLink.end());
  for(int cnt1=0;cnt1<rhs.facePLink.size();cnt1++){
    facePLink.push_back(rhs.facePLink[cnt1]);
  }
  for(int cnt1=0;cnt1<rhs.edgePLink.size();cnt1++){
    edgePLink.push_back(rhs.edgePLink[cnt1]);
  }  
}

// Copy from Vector3d
Vertex::Vertex(const Vector3d& rhs):Vector3d(rhs){
  facePLink.erase(facePLink.begin(),facePLink.end());
  edgePLink.erase(edgePLink.begin(),edgePLink.end());
  weight=0.5;
}

// Destructor
Vertex::~Vertex(){
  //this->Vector3d::~Vector3d();
}
 
// This is buggy code
const Vertex&
Vertex::operator=(const Vertex& rhs){
  if (this==&rhs)
  return *this;

  Vector3d::operator=(rhs);
  //edgePLink=rhs.edgePLink;
  //facePLink=rhs.facePLink;
  facePLink.erase(facePLink.begin(),facePLink.end());
  edgePLink.erase(edgePLink.begin(),edgePLink.end());
  for(int cnt1=0;cnt1<rhs.facePLink.size();cnt1++){
    facePLink.push_back(rhs.facePLink[cnt1]);
  }
  for(int cnt1=0;cnt1<rhs.edgePLink.size();cnt1++){
    edgePLink.push_back(rhs.edgePLink[cnt1]);
  }  
  vertexPointLink=rhs.vertexPointLink;
  weight=rhs.weight;
  return *this;
}

void
Vertex::SetV(Vertex& des){
  coord[0]=des.coord[0];
  coord[1]=des.coord[1];
  coord[2]=des.coord[2];
  weight=des.weight;
}

// Draw Point on GL window
// This should be called in glBegin(GL_POINTS)
void
Vertex::Draw(){
  glVertex3dv(coord); 
}

void
Vertex::Info(){
  Print();
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    cout << "edgePLink[" << cnt1 << "]==" << edgePLink[cnt1] << endl; 
  }
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    cout << "facePLink[" << cnt1 << "]==" << facePLink[cnt1] << endl; 
  }
  cout << "vertexPointLink==" << vertexPointLink << endl;
  cout << "weight==" << weight << endl;
}


// Calculate Vertex Point
// Q,R,comment from 
// "Recursively generated B-Spline surfaces on arbitrary topological meshes"
Vertex
Vertex::CalcVertexPoint(){
  //Vertex tmp1,tmp2,tmp3,tmp4;
  Vertex tmp1,tmp4;
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    tmp1.Plus(face[facePLink[cnt1]].ReturnFacePoint());
  }
  tmp1.Multi(1.0/facePLink.size()); // Q
  
  /*
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    tmp2.Plus(edge[edgePLink[cnt1]].ReturnEdgePoint());
  }
  tmp2.Multi(2.0/edgePLink.size()); // 2R
  tmp3=*this; // S

  tmp1.Plus(tmp2);
  tmp3.Multi((double)((int)edgePLink.size()-3));
  tmp1.Plus(tmp3);
  tmp1.Multi(1.0/edgePLink.size()); // result
  return tmp1;
  */

  // experimental
  tmp1.Multi(1.0-weight);
  tmp4=*this;
  tmp4.Multi(weight);
  
  tmp1.Plus(tmp4);
  tmp1.weight=weight;
  return tmp1;
}

void
Vertex::ConnectToEdgePoint(){
  int edgePLinkSize=edgePLink.size();
  for(int cnt1=0;cnt1<edgePLinkSize;cnt1++){
    /*
#ifdef DEBUG
    if(edgePLink[cnt1]>=edge.size()){
      cout << "Strange!!!!! edgePLink[" << cnt1 << "]=="
	   << edgePLink[cnt1] << ">=edge.size()" << endl;
    }
#endif
    */
    AddEdge(edge[edgePLink[cnt1]].edgePointLink,vertexPointLink);
  }  
}

// Set Edge Link
void
Vertex::SetEdgeLink(){
  edgePLink.push_back(edge.size()-1); // edge number
}

// Set Face Link
void
Vertex::SetFaceLink(){
  facePLink.push_back(face.size()-1);
}

void
Vertex::UnSetFaceLink(int num){
#ifdef DEBUG
  cout << "UnSetFaceLink Face[" << num << "]" << " in Vertex" << endl;
  cout << "facePLink" << endl;
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    cout << facePLink[cnt1] << " ";
  }
  cout << endl;
#endif
  //vector<int>::iterator end=remove(facePLink.begin(),facePLink.end(),num);
  //facePLink.erase(end,facePLink.end());  
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    if(facePLink[cnt1]==num){
      facePLink.erase(&facePLink[cnt1]);
      return;
    }
  }
#ifdef DEBUG
  cout << "UnSetFaceLink Face[" << num << "]" << endl;
  cout << "facePLink" << endl;
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    cout << facePLink[cnt1] << " ";
  }
  cout << endl;
#endif
}

void
Vertex::UnSetEdgeLink(int num){
#ifdef DEBUG
  cout << endl << "UnSetEdgeLink Edge[" << num << "]" << " in Vertex" << endl;
  cout << "edgePLink" << endl;
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    cout << edgePLink[cnt1] << " ";
  }
  cout << endl;
#endif
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    if(edgePLink[cnt1]==num){
      edgePLink.erase(&edgePLink[cnt1]);
      return;
    }
  }
#ifdef DEBUG
  cout << "UnSetEdgeLink Edge[" << num << "]" << endl;
  cout << "edgePLink" << endl;
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    cout << edgePLink[cnt1] << " ";
  }
  cout << endl;
#endif
}

void
Vertex::Sharp(){
  weight=1;
}

void
Vertex::Smooth(){
  weight=0;
}

void
Vertex::MoreSharp(){
  weight+=0.1;
  if(weight>1.0){
    weight=1.0;
  }
}

void
Vertex::MoreSmooth(){
  weight-=0.1;
  if(weight<0){
    weight=0;
  }
}

void
Vertex::SetWeight(float w){
  weight=w;
}

float
Vertex::ReturnWeight(){
  return weight;
} 

void
Vertex::WriteFile(FILE* fp){
  fprintf(fp,"%f %f %f %d %d %f\n",X(),Y(),Z(),
	  edgePLink.size(),facePLink.size(),weight);
  // edgePLink[]
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    fprintf(fp,"%d ",edgePLink[cnt1]);
  }
  // facePLink[]
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    fprintf(fp,"%d ",facePLink[cnt1]);
  }  
  fprintf(fp,"\n");
}

void
Vertex::ReadFile(FILE* fp){
  int edgePLinkSize,facePLinkSize,tmpEdge,tmpFace;
  fscanf(fp,"%lf %lf %lf %d %d %f",
	 &coord[0],&coord[1],&coord[2],&edgePLinkSize,&facePLinkSize,&weight);
  // check min,max
  if(coord[0]>maxX){
    maxX=coord[0];
  }else if(coord[0]<minX){
    minX=coord[0];
  }
  if(coord[1]>maxY){
    maxY=coord[1];
  }else if(coord[1]<minY){
    minY=coord[1];
  }
  if(coord[2]>maxZ){
    maxZ=coord[2];
  }else if(coord[2]<minZ){
    minZ=coord[2];
  }  
  
  for(int cnt1=0;cnt1<edgePLinkSize;cnt1++){
    fscanf(fp,"%d",&tmpEdge);
    edgePLink.push_back(tmpEdge+edgeOffset);
  }
  for(int cnt1=0;cnt1<facePLinkSize;cnt1++){
    fscanf(fp,"%d",&tmpFace);
    facePLink.push_back(tmpFace+faceOffset);
  }
  vertexPointLink=-1;
}

int
Vertex::HaveEdge(int num){
  for(int cnt1=0;cnt1<edgePLink.size();cnt1++){
    if(edgePLink[cnt1]==num){
      return 1; // have
    }
  }
  return 0; // not have
}
